// Copyright 2018 Espressif Systems (Shanghai) PTE LTD
// All rights reserved.

#include <string.h>
#include "esp_log.h"
#include "audio_error.h"
#include "audio_mem.h"
#include "audio_element.h"
#include "nullpipe.h"
#include "audio_type_def.h"
static const char *TAG = "NULLPIPE";

#define BUF_SIZE (128)

typedef struct nullpipe {
    int  samplerate;
    int  channel;
    unsigned char *buf;
    int  byte_num;
    int  at_eof;
} nullpipe_t;

typedef union {
	short audiosample16;
	struct audiosample16_bytes{
		unsigned h	:8;
		unsigned l	:8;
	} audiosample16_bytes;
} audiosample16_t;

#ifdef DEBUG_NULLPIPE_ENC_ISSUE
static FILE *infile;
#endif

static esp_err_t is_valid_nullpipe_samplerate(int samplerate)
{
    if ((samplerate != 11025)
        && (samplerate != 22050)
        && (samplerate != 44100)
        && (samplerate != 48000)) {
        ESP_LOGE(TAG, "The sample rate should be only 11025Hz, 22050Hz, 44100Hz, 48000Hz, here is %dHz. (line %d)", samplerate, __LINE__);
        return ESP_ERR_INVALID_ARG;
    }
    return ESP_OK;
}

static esp_err_t is_valid_nullpipe_channel(int channel)
{
    if ((channel != 1)
        && (channel != 2)) {
        ESP_LOGE(TAG, "The number of channels should be only 1 or 2, here is %d. (line %d)", channel, __LINE__);
        return ESP_ERR_INVALID_ARG;
    }
    return ESP_OK;
}

esp_err_t nullpipe_set_info(audio_element_handle_t self, int rate, int ch)
{
    nullpipe_t *nullpipe = (nullpipe_t *)audio_element_getdata(self);
    if (nullpipe->samplerate == rate && nullpipe->channel == ch) {
        return ESP_OK;
    }
    if ((is_valid_nullpipe_samplerate(rate) != ESP_OK)
        || (is_valid_nullpipe_channel(ch) != ESP_OK)) {
        return ESP_ERR_INVALID_ARG;
    } else {
        ESP_LOGI(TAG, "The reset sample rate and channel of audio stream are %d %d.", rate, ch);
    
        nullpipe->samplerate = rate;
        nullpipe->channel = ch;
        
    }
    return ESP_OK;
}

static esp_err_t nullpipe_destroy(audio_element_handle_t self)
{
    nullpipe_t *nullpipe = (nullpipe_t *)audio_element_getdata(self);
    audio_free(nullpipe);
    return ESP_OK;
}

static esp_err_t nullpipe_open(audio_element_handle_t self)
{
#ifdef NULLPIPE_MEMORY_ANALYSIS
    AUDIO_MEM_SHOW(TAG);
#endif
    ESP_LOGD(TAG, "nullpipe_open");
    nullpipe_t *nullpipe = (nullpipe_t *)audio_element_getdata(self);
    audio_element_info_t info = {0};
    audio_element_getinfo(self, &info);
    if (info.sample_rates
        && info.channels) {
        nullpipe->samplerate = info.sample_rates;
        nullpipe->channel = info.channels;
    }
    nullpipe->at_eof = 0;
    if (is_valid_nullpipe_samplerate(nullpipe->samplerate) != ESP_OK
        || is_valid_nullpipe_channel(nullpipe->channel) != ESP_OK) {
        return ESP_ERR_INVALID_ARG;
    }
    nullpipe->buf = (unsigned char *)calloc(1, BUF_SIZE);
    if (nullpipe->buf == NULL) {
        ESP_LOGE(TAG, "calloc buffer failed. (line %d)", __LINE__);
        return ESP_ERR_NO_MEM;
    }
    memset(nullpipe->buf, 0, BUF_SIZE);

#ifdef DEBUG_NULLPIPE_ENC_ISSUE
    char fileName[100] = {'//', 's', 'd', 'c', 'a', 'r', 'd', '//', 't', 'e', 's', 't', '.', 'p', 'c', 'm', '\0'};
    infile = fopen(fileName, "rb");
    if (!infile) {
        perror(fileName);
        return ESP_FAIL;
    }
#endif

    return ESP_OK;
}

static esp_err_t nullpipe_close(audio_element_handle_t self)
{
    ESP_LOGD(TAG, "nullpipe_close");
    nullpipe_t *nullpipe = (nullpipe_t *)audio_element_getdata(self);
    if(nullpipe->buf == NULL){
        audio_free(nullpipe->buf);
        nullpipe->buf = NULL;
    }   

#ifdef NULLPIPE_MEMORY_ANALYSIS
    AUDIO_MEM_SHOW(TAG);
#endif
#ifdef DEBUG_NULLPIPE_ENC_ISSUE
    fclose(infile);
#endif

    return ESP_OK;
}

static int nullpipe_process(audio_element_handle_t self, char *in_buffer, int in_len)
{
    nullpipe_t *nullpipe = (nullpipe_t *)audio_element_getdata(self);
    int ret = 0;

    int r_size = 0;    
    if (nullpipe->at_eof == 0) {
#ifdef DEBUG_NULLPIPE_ENC_ISSUE
        r_size = fread((char *)nullpipe->buf, 1, BUF_SIZE, infile);
#else
        r_size = audio_element_input(self, (char *)nullpipe->buf, BUF_SIZE);
#endif
    }
    if (r_size > 0) {
        if (r_size != BUF_SIZE) {
            nullpipe->at_eof = 1;
        }
        nullpipe->byte_num += r_size;

        unsigned char *pbuf = nullpipe->buf;
        
        for(int i=0;i<r_size;i+=2){
            audiosample16_t audiosample;
            audiosample.audiosample16_bytes.h = *pbuf;
            audiosample.audiosample16_bytes.l = *(pbuf+1);

            // process audio samples here e.g. scale by 1/16 (attenuation):
            // audiosample.audiosample16 = audiosample.audiosample16>>4;

            *pbuf = audiosample.audiosample16_bytes.h;
            *(pbuf+1) = audiosample.audiosample16_bytes.l;
            pbuf+=2;
        }
        ret = audio_element_output(self, (char *)nullpipe->buf, BUF_SIZE);
    } else {
        ret = r_size;
    }
    return ret;
}

audio_element_handle_t nullpipe_init(nullpipe_cfg_t *config)
{
    if (config == NULL) {
        ESP_LOGE(TAG, "config is NULL. (line %d)", __LINE__);
        return NULL;
    }
    nullpipe_t *nullpipe = audio_calloc(1, sizeof(nullpipe_t));
    AUDIO_MEM_CHECK(TAG, nullpipe, return NULL);     
    if (nullpipe == NULL) {
        ESP_LOGE(TAG, "audio_calloc failed for nullpipe. (line %d)", __LINE__);
        return NULL;
    }
    audio_element_cfg_t cfg = DEFAULT_AUDIO_ELEMENT_CONFIG();
    cfg.destroy = nullpipe_destroy;
    cfg.process = nullpipe_process;
    cfg.open = nullpipe_open;
    cfg.close = nullpipe_close;
    cfg.buffer_len = 0;
    cfg.tag = "nullpipe";
    cfg.task_stack = config->task_stack;
    cfg.task_prio = config->task_prio;
    cfg.task_core = config->task_core;
    cfg.out_rb_size = config->out_rb_size;
    audio_element_handle_t el = audio_element_init(&cfg);
    AUDIO_MEM_CHECK(TAG, el, {audio_free(nullpipe); return NULL;});
    nullpipe->samplerate = config->samplerate;
    nullpipe->channel = config->channel;
    audio_element_setdata(el, nullpipe);
    audio_element_info_t info = {0};
    audio_element_setinfo(el, &info);
    ESP_LOGD(TAG, "nullpipe_init");
    return el;
}
